@@ -12,12 +12,13 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-from struct import unpack, pack
+from struct import pack, Struct
 from string import printable
 
 from .header import Header, DTB_PROP, DTB_BEGIN_NODE, DTB_END_NODE
 from .misc import is_string, line_offset
 
+BIGENDIAN_WORD = Struct(">I")
 
 ########################################################################################################################
 # Helper methods
@@ -41,11 +42,10 @@ def new_property(name: str, raw_value: bytes) -> object:
     elif len(raw_value) and len(raw_value) % 4 == 0:
         obj = PropWords(name)
         # Extract words from raw value
-        for i in range(0, len(raw_value), 4):
-            obj.append(unpack(">I", raw_value[i:i + 4])[0])
+        obj.data = [BIGENDIAN_WORD.unpack(raw_value[i:i + 4])[0] for i in range(0, len(raw_value), 4)]
         return obj
 
-    elif len(raw_value) and len(raw_value):
+    elif len(raw_value):
         return PropBytes(name, data=raw_value)
 
     else:
@@ -62,6 +62,10 @@ class BaseItem:
     def name(self):
         return self._name
 
+    @property
+    def label(self):
+        return self._label
+
     @property
     def parent(self):
         return self._parent
@@ -85,6 +89,7 @@ class BaseItem:
         assert isinstance(name, str)
         assert all(c in printable for c in name), "The value must contain just printable chars !"
         self._name = name
+        self._label = None
         self._parent = None
 
     def __str__(self):
@@ -101,6 +106,17 @@ class BaseItem:
         assert all(c in printable for c in value), "The value must contain just printable chars !"
         self._name = value
 
+    def set_label(self, value: str):
+        """ 
+        Set item label
+        
+        :param value: The label in string format
+        """
+        assert isinstance(value, str)
+        assert all(c in printable for c in value), "The value must contain just printable chars !"
+        self._label = value
+
+
     def set_parent(self, value):
         """ 
         Set item parent
@@ -342,8 +358,7 @@ class PropWords(Property):
             strpos = len(strings)
             strings += self.name + '\0'
         blob = pack('>III', DTB_PROP, len(self.data) * 4, strpos)
-        for word in self.data:
-            blob += pack('>I', word)
+        blob += bytes().join([BIGENDIAN_WORD.pack(word) for word in self.data])
         pos += len(blob)
         return blob, strings, pos
 
@@ -708,8 +723,16 @@ class Node(BaseItem):
         :param tabsize: Tabulator size in count of spaces
         :param depth: Start depth for line
         """
-        dts  = line_offset(tabsize, depth, self.name + ' {\n')
-        dts += ''.join(prop.to_dts(tabsize, depth + 1) for prop in self._props)
+        if self._label is not None:
+            dts  = line_offset(tabsize, depth, self._label + ': ' + self.name + ' {\n')
+        else:
+            dts  = line_offset(tabsize, depth, self.name + ' {\n')
+        # phantom properties which maintain reference state info
+        # have names ending with _with_references
+        # don't write those out to dts file
+        dts += ''.join(
+            prop.to_dts(tabsize, depth + 1) 
+            for prop in self._props if prop.name.endswith('_with_references') is False)
         dts += ''.join(node.to_dts(tabsize, depth + 1) for node in self._nodes)
         dts += line_offset(tabsize, depth, "};\n")
         return dts
@@ -731,8 +754,11 @@ class Node(BaseItem):
             blob += pack('b', 0) * (4 - (len(blob) % 4))
         pos += len(blob)
         for prop in self._props:
-            (data, strings, pos) = prop.to_dtb(strings, pos, version)
-            blob += data
+            # phantom property too maintain reference state should
+            # not write out to dtb file
+            if prop.name.endswith('_with_references') is False:
+                (data, strings, pos) = prop.to_dtb(strings, pos, version)
+                blob += data
         for node in self._nodes:
             (data, strings, pos) = node.to_dtb(strings, pos, version)
             blob += data
